import 'dart:math' as math;
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';
import 'package:sheet/src/snap_scroll_controller.dart';
import 'package:sheet/src/widgets/resizable_sheet.dart';
import 'package:sheet/src/widgets/scroll_to_top_status_handler.dart';

import '../sheet.dart';

enum SheetFit { loose, expand }

class Sheet extends StatelessWidget {
  const Sheet({
    Key? key,
    required this.child,
    this.controller,
    this.physics,
    this.initialExtent,
    this.minExtent,
    this.maxExtent,
    this.minInteractionExtent = 20.0,
    this.backgroundColor,
    this.clipBehavior,
    this.shape,
    this.elevation,
    this.fit = SheetFit.loose,
    this.resizable = false,
    this.padding = EdgeInsets.zero,
    this.minResizableExtent,
  })  : raw = false,
        super(key: key);

  const Sheet.raw({
    Key? key,
    required this.child,
    this.controller,
    this.physics,
    this.initialExtent,
    this.minExtent,
    this.maxExtent,
    this.minInteractionExtent = 20.0,
    this.fit = SheetFit.loose,
    this.resizable = false,
    this.padding = EdgeInsets.zero,
    this.minResizableExtent,
  })  : raw = true,
        shape = null,
        elevation = null,
        backgroundColor = null,
        clipBehavior = null,
        super(key: key);

  final Widget child;
  final SheetPhysics? physics;
  final SheetController? controller;

  final EdgeInsets padding;

  final double? initialExtent;

  final double? minExtent;
  final double? maxExtent;
  final double minInteractionExtent;

  final bool resizable;
  final SheetFit fit;

  final bool raw;

  final double? minResizableExtent;

  /// The state from the closest instance of this class that encloses the given context.
  ///
  /// Typical usage is as follows:
  ///
  /// ```dart
  /// ScrollableState scrollable = Scrollable.of(context);
  /// ```
  ///
  /// Calling this method will create a dependency on the closest [SheetScrollable]
  /// in the [context], if there is one.
  static SheetState? of(BuildContext context) {
    return SheetScrollable.of(context);
  }

  /// The bottom sheet's background color.
  ///
  /// Defines the bottom sheet's [Material.color].
  ///
  /// Defaults to null and falls back to [Material]'s default.
  final Color? backgroundColor;

  /// The z-coordinate at which to place this material relative to its parent.
  ///
  /// This controls the size of the shadow below the material.
  ///
  /// Defaults to 0. The value is non-negative.
  final double? elevation;

  /// The shape of the bottom sheet.
  ///
  /// Defines the bottom sheet's [Material.shape].
  ///
  /// Defaults to null and falls back to [Material]'s default.
  final ShapeBorder? shape;

  /// {@macro flutter.material.Material.clipBehavior}
  ///
  /// Defines the bottom sheet's [Material.clipBehavior].
  ///
  /// Use this property to enable clipping of content when the bottom sheet has
  /// a custom [shape] and the content can extend past this shape. For example,
  /// a bottom sheet with rounded corners and an edge-to-edge [Image] at the
  /// top.
  ///
  /// If this property is null then [BottomSheetThemeData.clipBehavior] of
  /// [ThemeData.bottomSheetTheme] is used. If that's null then the behavior
  /// will be [Clip.none].
  final Clip? clipBehavior;

  @override
  Widget build(BuildContext context) {
    Widget sheet;

    if (raw) {
      sheet = child;
    } else {
      final BottomSheetThemeData bottomSheetTheme = Theme.of(context).bottomSheetTheme;
      final Color? color = backgroundColor ?? bottomSheetTheme.backgroundColor;
      final double elevation = this.elevation ?? bottomSheetTheme.elevation ?? 0;
      final ShapeBorder? shape = this.shape ?? bottomSheetTheme.shape;
      final Clip clipBehavior = this.clipBehavior ?? bottomSheetTheme.clipBehavior ?? Clip.none;

      sheet = Material(
        color: color,
        elevation: elevation,
        shape: shape,
        clipBehavior: clipBehavior,
        child: child,
      );
    }

    return SheetScrollable(
      initialExtent: initialExtent?.clamp(minExtent ?? 0, maxExtent ?? double.infinity),
      minInteractionExtent: minInteractionExtent,
      physics: physics,
      controller: controller,
      axisDirection: AxisDirection.down,
      scrollBehavior: SheetBehaviour(),
      viewportBuilder: (BuildContext context, ViewportOffset offset) {
        return DefaultSheetScrollController(
          child: ScrollToTopStatusBarHandler(
            child: SheetViewport(
              clipBehavior: Clip.antiAlias,
              axisDirection: AxisDirection.down,
              offset: offset,
              minExtent: minExtent,
              maxExtent: maxExtent,
              fit: fit,
              child: Padding(
                padding: padding,
                child: ResizableSheetChild(
                  resizable: resizable,
                  offset: offset,
                  minExtent: minResizableExtent ?? 0,
                  child: sheet,
                ),
              ),
            ),
          ),
        );
      },
    );
  }
}

class DefaultSheetScrollController extends StatefulWidget {
  const DefaultSheetScrollController({Key? key, required this.child}) : super(key: key);

  final Widget child;
  @override
  _DefaultSheetScrollControllerState createState() => _DefaultSheetScrollControllerState();
}

class _DefaultSheetScrollControllerState extends State<DefaultSheetScrollController> {
  SheetPrimaryScrollController _controller = SheetPrimaryScrollController(sheetPosition: null);
  ScrollPosition? _position;

  @override
  void dispose() {
    _controller.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return PrimaryScrollController(
      controller: _controller,
      child: widget.child,
    );
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    final SheetPosition position = SheetScrollable.of(context)!.position;
    if (_position != position) {
      _controller.dispose();
      _position = position;
      _controller = SheetPrimaryScrollController(sheetPosition: position);
      setState(() {});
    }
  }
}

/// A [ScrollController] suitable for use in a [SheetPosition] created
/// by a [Sheet].
///
/// If a [Sheet] contains content that is exceeds the height
/// of its container, this controller will allow the sheet to both be dragged to
/// fill the container and then scroll the child content.
///
/// See also:
///
///  * [SheetPosition], which manages the positioning logic for
///    this controller.
///  * [PrimarySheetController], which can be used to establish a
///    [_SheetScrollController] as the primary controller for
///    descendants.
class SheetController extends ScrollController {
  SheetController({
    String? debugLabel,
  }) : super(
          debugLabel: debugLabel,
          initialScrollOffset: 0,
        );

  final ProxyAnimation _animation = ProxyAnimation();
  Animation<double> get animation => _animation;

  void updateAnimation() {
    if (hasClients) {
      final ScrollPosition position = positions.first;
      if (position is SheetPosition) {
        _animation.parent = position.animation;
        return;
      }
    }
    _animation.parent = kAlwaysDismissedAnimation;
    return;
  }

  @override
  SheetPosition get position {
    return super.position as SheetPosition;
  }

  @override
  void attach(ScrollPosition position) {
    super.attach(position);
    updateAnimation();
  }

  @override
  void detach(ScrollPosition position) {
    super.detach(position);
    updateAnimation();
  }

  @override
  SheetPosition createScrollPosition(
    ScrollPhysics physics,
    ScrollContext context,
    ScrollPosition? oldPosition,
  ) {
    final double? initialPixels = (context is SheetContext) ? context.initialExtent : null;
    return SheetPosition(
      physics: physics,
      context: context,
      oldPosition: oldPosition,
      initialPixels: initialPixels ?? 0.0,
    );
  }

  Future<void> relativeAnimateTo(double offset,
      {required Duration duration, required Curve curve}) async {
    assert(positions.isNotEmpty, 'ScrollController not attached to any scroll views.');
    await Future.wait<void>(<Future<void>>[
      for (final ScrollPosition position in positions)
        (position as SheetPosition).relativeAnimateTo(offset, duration: duration, curve: curve),
    ]);
  }

  void relativeJumpTo(double offset) {
    assert(positions.isNotEmpty, 'ScrollController not attached to any scroll views.');
    for (final ScrollPosition position in positions)
      (position as SheetPosition).relativeJumpTo(offset);
  }

  @override
  void debugFillDescription(List<String> description) {
    super.debugFillDescription(description);
  }
}

/// A scroll position that manages scroll activities for
/// [_SheetScrollController].
///
/// This class is a concrete subclass of [ScrollPosition] logic that handles a
/// single [ScrollContext], such as a [Scrollable]. An instance of this class
/// manages [ScrollActivity] instances, which changes the
/// [SheetPosition.currentExtent] or visible content offset in the
/// [Scrollable]'s [Viewport]
///
/// See also:
///
///  * [_SheetScrollController], which uses this as its [ScrollPosition].
class SheetPosition extends ScrollPositionWithSingleContext {
  SheetPosition({
    required ScrollPhysics physics,
    required ScrollContext context,
    double initialPixels = 0.0,
    bool keepScrollOffset = true,
    ScrollPosition? oldPosition,
    String? debugLabel,
  }) : super(
          physics: physics,
          context: context,
          initialPixels: initialPixels,
          keepScrollOffset: keepScrollOffset,
          oldPosition: oldPosition,
          debugLabel: debugLabel,
        );

  bool preventingPop = false;

  Future<void> relativeAnimateTo(double to, {required Duration duration, required Curve curve}) {
    assert(to >= 0 && to <= 1);
    return super.animateTo(to * maxScrollExtent, duration: duration, curve: curve);
  }

  @override
  Future<void> animateTo(double to, {required Duration duration, required Curve curve}) {
    return super
        .animateTo(to.clamp(minScrollExtent, maxScrollExtent), duration: duration, curve: curve);
  }

  void relativeJumpTo(double to) {
    assert(to >= 0 && to <= 1);
    return super.jumpTo(to * maxScrollExtent);
  }

  @override
  void jumpTo(double value) {
    return super.jumpTo(value.clamp(minScrollExtent, maxScrollExtent));
  }

  late final AnimationController _controller = AnimationController(vsync: context.vsync);
  Animation<double> get animation => _controller;
  @override
  double setPixels(double newPixels) {
    _controller.value = (newPixels / maxScrollExtent).clamp(0, 1);
    return super.setPixels(newPixels);
  }

  @override
  void forcePixels(double value) {
    _controller.value = (value / maxScrollExtent).clamp(0, 1);
    super.forcePixels(value);
  }

  @override
  void dispose() {
    _controller.dispose();
    super.dispose();
  }
}

class SheetViewport extends SingleChildRenderObjectWidget {
  const SheetViewport({
    Key? key,
    this.axisDirection = AxisDirection.down,
    required this.offset,
    this.minExtent,
    this.maxExtent,
    Widget? child,
    required this.fit,
    required this.clipBehavior,
  })  : assert(axisDirection != null),
        assert(clipBehavior != null),
        super(key: key, child: child);

  final AxisDirection axisDirection;
  final ViewportOffset offset;
  final Clip clipBehavior;
  final double? minExtent;
  final double? maxExtent;
  final SheetFit fit;

  @override
  _RenderSheetViewport createRenderObject(BuildContext context) {
    return _RenderSheetViewport(
      axisDirection: axisDirection,
      offset: offset,
      clipBehavior: clipBehavior,
      minExtent: minExtent,
      maxExtent: maxExtent,
      fit: fit,
    );
  }

  @override
  void updateRenderObject(BuildContext context, _RenderSheetViewport renderObject) {
    // Order dependency: The offset setter reads the axis direction.
    renderObject
      ..axisDirection = axisDirection
      ..offset = offset
      ..clipBehavior = clipBehavior
      ..minExtent = minExtent
      ..maxExtent = maxExtent
      ..fit = fit;
  }
}

class _RenderSheetViewport extends RenderBox
    with RenderObjectWithChildMixin<RenderBox>
    implements RenderAbstractViewport {
  _RenderSheetViewport({
    AxisDirection axisDirection = AxisDirection.down,
    required ViewportOffset offset,
    double cacheExtent = RenderAbstractViewport.defaultCacheExtent,
    RenderBox? child,
    required Clip clipBehavior,
    SheetFit fit = SheetFit.expand,
    double? minExtent,
    double? maxExtent,
  })  : assert(axisDirection != null),
        assert(offset != null),
        assert(cacheExtent != null),
        assert(clipBehavior != null),
        _axisDirection = axisDirection,
        _offset = offset,
        _fit = fit,
        _minExtent = minExtent,
        _maxExtent = maxExtent,
        _cacheExtent = cacheExtent,
        _clipBehavior = clipBehavior {
    this.child = child;
  }

  AxisDirection get axisDirection => _axisDirection;
  AxisDirection _axisDirection;
  set axisDirection(AxisDirection value) {
    assert(value != null);
    if (value == _axisDirection) return;
    _axisDirection = value;
    markNeedsLayout();
  }

  Axis get axis => axisDirectionToAxis(axisDirection);

  ViewportOffset get offset => _offset;
  ViewportOffset _offset;
  set offset(ViewportOffset value) {
    assert(value != null);
    if (value == _offset) return;
    if (attached) _offset.removeListener(_hasScrolled);
    _offset = value;
    if (attached) _offset.addListener(_hasScrolled);
    markNeedsLayout();
  }

  /// {@macro flutter.rendering.RenderViewportBase.cacheExtent}
  double get cacheExtent => _cacheExtent;
  double _cacheExtent;
  set cacheExtent(double value) {
    assert(value != null);
    if (value == _cacheExtent) return;
    _cacheExtent = value;
    markNeedsLayout();
  }

  /// {@macro flutter.rendering.RenderViewportBase.cacheExtent}
  SheetFit get fit => _fit;
  SheetFit _fit;
  set fit(SheetFit value) {
    assert(value != null);
    if (value == _fit) return;
    _fit = value;
    markNeedsLayout();
  }

  /// {@macro flutter.material.Material.clipBehavior}
  ///
  /// Defaults to [Clip.none], and must not be null.
  Clip get clipBehavior => _clipBehavior;
  Clip _clipBehavior = Clip.none;
  set clipBehavior(Clip value) {
    assert(value != null);
    if (value != _clipBehavior) {
      _clipBehavior = value;
      markNeedsPaint();
      markNeedsSemanticsUpdate();
    }
  }

  void _hasScrolled() {
    if (offset.pixels > _maxScrollExtent) {
      _maxScrollExtentDuringOverflow ??= _maxScrollExtent;
      _childExtentDuringOverflow ??= child!.size.height;
      markNeedsLayout();
    } else {
      _maxScrollExtentDuringOverflow = null;
      _childExtentDuringOverflow = null;
    }
    markNeedsPaint();
    markNeedsSemanticsUpdate();
  }

  @override
  void setupParentData(RenderObject child) {
    // We don't actually use the offset argument in BoxParentData, so let's
    // avoid allocating it at all.
    if (child.parentData is! ParentData) child.parentData = ParentData();
  }

  @override
  void attach(PipelineOwner owner) {
    super.attach(owner);
    _offset.addListener(_hasScrolled);
  }

  @override
  void detach() {
    _offset.removeListener(_hasScrolled);
    super.detach();
  }

  @override
  bool get isRepaintBoundary => true;

  double? get minExtent => _minExtent;
  double? _minExtent;
  set minExtent(double? value) {
    if (value != _minExtent) {
      _minExtent = value;
      markNeedsLayout();
    }
  }

  double? get maxExtent => _maxExtent;
  double? _maxExtent;
  set maxExtent(double? value) {
    if (value != _maxExtent) {
      _maxExtent = value;
      markNeedsLayout();
    }
  }

  double get _viewportExtent {
    assert(hasSize);
    switch (axis) {
      case Axis.horizontal:
        return size.width;
      case Axis.vertical:
        return size.height;
    }
  }

  double get _minScrollExtent {
    assert(hasSize);
    return minExtent ?? 0.0;
  }

  double get _maxScrollExtent {
    assert(hasSize);
    if (_maxScrollExtentDuringOverflow != null) return _maxScrollExtentDuringOverflow!;
    if (child == null) return 0.0;
    switch (axis) {
      case Axis.horizontal:
        return math.max(0.0, child!.size.width - size.width);
      case Axis.vertical:
        return math.max(0.0, child!.size.height);
    }
  }

  double? _maxScrollExtentDuringOverflow;
  double? _childExtentDuringOverflow;

  BoxConstraints _getInnerConstraints(BoxConstraints constraints) {
    switch (axis) {
      case Axis.horizontal:
        return constraints.heightConstraints();
      case Axis.vertical:
        return constraints.widthConstraints();
    }
  }

  @override
  double computeMinIntrinsicWidth(double height) {
    return constraints.maxHeight;
  }

  @override
  double computeMaxIntrinsicWidth(double height) {
    return constraints.maxWidth;
  }

  @override
  double computeMinIntrinsicHeight(double width) {
    return constraints.maxHeight;
  }

  @override
  double computeMaxIntrinsicHeight(double width) {
    return constraints.maxHeight;
  }

  // We don't override computeDistanceToActualBaseline(), because we
  // want the default behavior (returning null). Otherwise, as you
  // scroll, it would shift in its parent if the parent was baseline-aligned,
  // which makes no sense.

  @override
  Size computeDryLayout(BoxConstraints constraints) {
    if (child == null) {
      return constraints.smallest;
    }
    final Size childSize = child!.getDryLayout(_getInnerConstraints(constraints));
    return constraints.biggest;
  }

  @override
  void performLayout() {
    final BoxConstraints constraints = this.constraints;
    if (child == null) {
      size = constraints.smallest;
    } else {
      final double? minHeight = _maxScrollExtentDuringOverflow != null
          ? _childExtentDuringOverflow! + offset.pixels - _maxScrollExtentDuringOverflow!
          : null;
      final bool expand = fit == SheetFit.expand;
      //  child!.layout(_getInnerConstraints(constraints), parentUsesSize: true);
      double height = maxExtent ?? constraints.biggest.height;
      height = math.max(height, offset.pixels);
      if (minHeight == null) {
        height = height.clamp(0, constraints.maxHeight);
      } else {
        height = math.max(height, minHeight);
      }

      final BoxConstraints newContstraints = BoxConstraints(
        minHeight: minHeight ?? (expand ? height : 0),
        maxHeight: height,
        minWidth: constraints.minWidth,
        maxWidth: constraints.maxWidth,
      );
      child!.layout(newContstraints, parentUsesSize: true);
      size = constraints.biggest;
    }

    offset.applyViewportDimension(_viewportExtent);
    offset.applyContentDimensions(_minScrollExtent, _maxScrollExtent);
  }

  Offset get _paintOffset => _paintOffsetForPosition(offset.pixels);

  Offset _paintOffsetForPosition(double position) {
    assert(axisDirection != null);
    switch (axisDirection) {
      case AxisDirection.up:
        return Offset(0.0, position - child!.size.height + size.height);
      case AxisDirection.down:
        return Offset(0.0, -position + size.height);
      case AxisDirection.left:
        return Offset(position - child!.size.width + size.width, 0.0);
      case AxisDirection.right:
        return Offset(-position, 0.0);
    }
  }

  bool _shouldClipAtPaintOffset(Offset paintOffset) {
    assert(child != null);
    return paintOffset.dx < 0 ||
        paintOffset.dy < 0 ||
        paintOffset.dx + child!.size.width > size.width ||
        paintOffset.dy + child!.size.height > size.height;
  }

  @override
  void paint(PaintingContext context, Offset offset) {
    if (child != null) {
      final Offset paintOffset = _paintOffset;

      void paintContents(PaintingContext context, Offset offset) {
        context.paintChild(child!, offset + paintOffset);
      }

      if (_shouldClipAtPaintOffset(paintOffset) && clipBehavior != Clip.none) {
        _clipRectLayer.layer = context.pushClipRect(
          needsCompositing,
          offset,
          Offset.zero & size,
          paintContents,
          clipBehavior: clipBehavior,
          oldLayer: _clipRectLayer.layer,
        );
      } else {
        _clipRectLayer.layer = null;
        paintContents(context, offset);
      }
    }
  }

  final LayerHandle<ClipRectLayer> _clipRectLayer = LayerHandle<ClipRectLayer>();

  @override
  void dispose() {
    _clipRectLayer.layer = null;
    super.dispose();
  }

  @override
  void applyPaintTransform(RenderBox child, Matrix4 transform) {
    final Offset paintOffset = _paintOffset;
    transform.translate(paintOffset.dx, paintOffset.dy);
  }

  @override
  Rect? describeApproximatePaintClip(RenderObject? child) {
    if (child != null && _shouldClipAtPaintOffset(_paintOffset)) return Offset.zero & size;
    return null;
  }

  @override
  bool hitTestChildren(BoxHitTestResult result, {required Offset position}) {
    if (child != null) {
      return result.addWithPaintOffset(
        offset: _paintOffset,
        position: position,
        hitTest: (BoxHitTestResult result, Offset? transformed) {
          assert(transformed == position + -_paintOffset);
          return child!.hitTest(result, position: transformed!);
        },
      );
    }
    return false;
  }

  @override
  RevealedOffset getOffsetToReveal(RenderObject target, double alignment, {Rect? rect}) {
    rect ??= target.paintBounds;
    if (target is! RenderBox) return RevealedOffset(offset: offset.pixels, rect: rect);

    final RenderBox targetBox = target;
    final Matrix4 transform = targetBox.getTransformTo(child);
    final Rect bounds = MatrixUtils.transformRect(transform, rect);
    final Size contentSize = child!.size;

    final double leadingScrollOffset;
    final double targetMainAxisExtent;
    final double mainAxisExtent;

    assert(axisDirection != null);
    switch (axisDirection) {
      case AxisDirection.up:
        mainAxisExtent = size.height;
        leadingScrollOffset = contentSize.height - bounds.bottom;
        targetMainAxisExtent = bounds.height;
        break;
      case AxisDirection.right:
        mainAxisExtent = size.width;
        leadingScrollOffset = bounds.left;
        targetMainAxisExtent = bounds.width;
        break;
      case AxisDirection.down:
        mainAxisExtent = size.height;
        leadingScrollOffset = bounds.top;
        targetMainAxisExtent = bounds.height;
        break;
      case AxisDirection.left:
        mainAxisExtent = size.width;
        leadingScrollOffset = contentSize.width - bounds.right;
        targetMainAxisExtent = bounds.width;
        break;
    }

    final double targetOffset =
        leadingScrollOffset - (mainAxisExtent - targetMainAxisExtent) * alignment;
    final Rect targetRect = bounds.shift(_paintOffsetForPosition(targetOffset));
    return RevealedOffset(offset: targetOffset, rect: targetRect);
  }

  @override
  void showOnScreen({
    RenderObject? descendant,
    Rect? rect,
    Duration duration = Duration.zero,
    Curve curve = Curves.ease,
  }) {
    if (!offset.allowImplicitScrolling) {
      return super.showOnScreen(
        descendant: descendant,
        rect: rect,
        duration: duration,
        curve: curve,
      );
    }

    final Rect? newRect = RenderViewportBase.showInViewport(
      descendant: descendant,
      viewport: this,
      offset: offset,
      rect: rect,
      duration: duration,
      curve: curve,
    );
    super.showOnScreen(
      rect: newRect,
      duration: duration,
      curve: curve,
    );
  }

  @override
  Rect describeSemanticsClip(RenderObject child) {
    assert(axis != null);
    switch (axis) {
      case Axis.vertical:
        return Rect.fromLTRB(
          semanticBounds.left,
          semanticBounds.top - cacheExtent,
          semanticBounds.right,
          semanticBounds.bottom + cacheExtent,
        );
      case Axis.horizontal:
        return Rect.fromLTRB(
          semanticBounds.left - cacheExtent,
          semanticBounds.top,
          semanticBounds.right + cacheExtent,
          semanticBounds.bottom,
        );
    }
  }
}
