import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';
import 'package:sheet/sheet.dart';


/// A [ScrollController] suitable for use in a [SheetPrimaryScrollPosition] created
/// by a [Sheet].
///
/// If a [Sheet] contains content that is exceeds the height
/// of its container, this controller will allow the sheet to both be dragged to
/// fill the container and then scroll the child content.
///
/// See also:
///
///  * [SheetPrimaryScrollPosition], which manages the positioning logic for
///    this controller.
///  * [PrimarySheetController], which can be used to establish a
///    [_SheetScrollController] as the primary controller for
///    descendants.
class SheetPrimaryScrollController extends ScrollController {
  SheetPrimaryScrollController({
    double initialScrollOffset = 0.0,
    String? debugLabel,
    required this.sheetPosition,
  }) : super(
          debugLabel: debugLabel,
          initialScrollOffset: initialScrollOffset,
        );

  final SheetPosition? sheetPosition;

  @override
  SheetPrimaryScrollPosition createScrollPosition(
    ScrollPhysics physics,
    ScrollContext context,
    ScrollPosition? oldPosition,
  ) {
    return SheetPrimaryScrollPosition(
      physics: physics,
      context: context,
      oldPosition: oldPosition,
      sheetPosition: sheetPosition ?? SheetPosition(physics: physics, context: context),
    );
  }

  @override
  void debugFillDescription(List<String> description) {
    super.debugFillDescription(description);
  }
}

/// A scroll position that manages scroll activities for
/// [_SheetScrollController].
///
/// This class is a concrete subclass of [ScrollPosition] logic that handles a
/// single [ScrollContext], such as a [Scrollable]. An instance of this class
/// manages [ScrollActivity] instances, which changes the
/// [SheetPrimaryScrollPosition.currentExtent] or visible content offset in the
/// [Scrollable]'s [Viewport]
///
/// See also:
///
///  * [_SheetScrollController], which uses this as its [ScrollPosition].
class SheetPrimaryScrollPosition extends ScrollPositionWithSingleContext {
  SheetPrimaryScrollPosition({
    required ScrollPhysics physics,
    required ScrollContext context,
    double initialPixels = 0.0,
    bool keepScrollOffset = true,
    required this.sheetPosition,
    ScrollPosition? oldPosition,
    String? debugLabel,
  }) : super(
          physics: physics,
          context: context,
          initialPixels: initialPixels,
          keepScrollOffset: keepScrollOffset,
          oldPosition: oldPosition,
          debugLabel: debugLabel,
        );

  final SheetPosition sheetPosition;

  @override
  void applyUserOffset(double delta) {
//    print('max ${sheetPosition.maxScrollExtent}, pix ${sheetPosition.pixels}, $delta');
    final bool sheetCanDrag = !sheetPosition.preventingPop && sheetPosition.maxScrollExtent > sheetPosition.pixels;
    final bool scrollReachedTop = pixels <= minScrollExtent && delta >= 0;

    if (sheetCanDrag || scrollReachedTop) {
      final double pixels = sheetPosition.pixels -
          sheetPosition.physics.applyPhysicsToUserOffset(sheetPosition, delta);
      sheetPosition
          .forcePixels(pixels.clamp(sheetPosition.minScrollExtent, sheetPosition.maxScrollExtent));
    } else {
      final ScrollDirection direction =
          delta > 0.0 ? ScrollDirection.forward : ScrollDirection.reverse;
      updateUserScrollDirection(direction);
      setPixels(pixels - physics.applyPhysicsToUserOffset(this, delta));
    }
  }

  double get _additionalMinExtent {
    return sheetPosition.hasPixels &&
            sheetPosition.hasContentDimensions &&
            sheetPosition.maxScrollExtent == sheetPosition.pixels
        ? 0.0
        : 1.0;
  }

  double get _additionalMaxExtent => sheetPosition.hasPixels &&
          sheetPosition.hasContentDimensions &&
          sheetPosition.hasPixels &&
          sheetPosition.maxScrollExtent == sheetPosition.pixels
      ? 0.0
      : 1.0;

  @override
  bool applyContentDimensions(double minScrollExtent, double maxScrollExtent) {
    // We need to provide some extra extent if we haven't yet reached the max or
    // min extents. Otherwise, a list with fewer children than the extent of
    // the available space will get stuck.
    return super.applyContentDimensions(
      minScrollExtent - _additionalMinExtent,
      maxScrollExtent + _additionalMaxExtent,
    );
  }

  @override
  void goBallistic(double velocity) {
    if (velocity > 0.0 && sheetPosition.maxScrollExtent == sheetPosition.pixels ||
        (velocity < 0.0 && pixels > 0)) {
      super.goBallistic(velocity);
      return;
    }
    if (sheetPosition.hasContentDimensions && !sheetPosition.preventingPop) {
      sheetPosition.goBallistic(velocity);
    }
    goIdle();
  }

  //@override
  //double get pixels => super.pixels + viewportDimension;
}
