// ignore_for_file: prefer_const_constructors

import 'package:flutter/material.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mocktail/mocktail.dart';
import 'package:sheet/src/widgets/min_interaction.dart';

import '../../default_test_value.dart';

Future<void> testMinInteractionZone({
  required WidgetTester tester,
  required AxisDirection direction,
  required double interationZoneExtent,
  required Offset tapOffset,
  required bool shouldDetectTap,
}) async {
  bool tapped = false;
  await tester.pumpWidget(
    Listener(
      onPointerDown: (_) {
        tapped = true;
      },
      child: MinInteractionZone(
        direction: direction,
        extent: interationZoneExtent,
        child: Container(
          width: double.infinity,
        ),
      ),
    ),
  );
  await tester.tapAt(tapOffset);
  expect(tapped, shouldDetectTap ? isTrue : isFalse);
}

extension on WidgetTester {
  Future<void> pumpInteractionZone({
    required AxisDirection direction,
    required double interationZoneExtent,
    required VoidCallback onTap,
  }) {
    return pumpWidget(
      GestureDetector(
        onTap: onTap,
        child: MinInteractionZone(
          direction: direction,
          extent: interationZoneExtent,
          child: Container(
            width: double.infinity,
          ),
        ),
      ),
    );
  }
}

void main() {
  late VoidCallback mockOnTap;

  setUp(() {
    mockOnTap = MockVoidCallback();
    when(mockOnTap).thenAnswer((_) {});
  });

  group('MinInteractionZone', () {
    group('with extent detects pointer', () {
      const double interationZoneExtent = 100.0;
      group('AxisDirection.down', () {
        const AxisDirection direction = AxisDirection.down;
        testWidgets('Pointer at top left', (WidgetTester tester) async {
          await tester.pumpInteractionZone(
            direction: direction,
            interationZoneExtent: interationZoneExtent,
            onTap: mockOnTap,
          );

          await tester.tapAt(Offset(0, 0));
          verify(mockOnTap).called(1);
        });

        testWidgets('Pointer at top right', (WidgetTester tester) async {
          await tester.pumpInteractionZone(
            direction: direction,
            interationZoneExtent: interationZoneExtent,
            onTap: mockOnTap,
          );

          await tester.tapAt(Offset(kScreenSize.width - 0.1, 0));
          verify(mockOnTap).called(1);
        });

        testWidgets('Pointer at center', (WidgetTester tester) async {
          await tester.pumpInteractionZone(
            direction: direction,
            interationZoneExtent: interationZoneExtent,
            onTap: mockOnTap,
          );

          await tester
              .tapAt(Offset(kScreenSize.width / 2, interationZoneExtent / 2));
          verify(mockOnTap).called(1);
        });
        testWidgets('Pointer at bottom left', (WidgetTester tester) async {
          await tester.pumpInteractionZone(
            direction: direction,
            interationZoneExtent: interationZoneExtent,
            onTap: mockOnTap,
          );
          await tester.tapAt(Offset(0, interationZoneExtent - 0.1));
          verify(mockOnTap).called(1);
        });

        testWidgets('Pointer at bottom right', (WidgetTester tester) async {
          await tester.pumpInteractionZone(
            direction: direction,
            interationZoneExtent: interationZoneExtent,
            onTap: mockOnTap,
          );
          await tester.tapAt(
              Offset(kScreenSize.width - 0.1, interationZoneExtent - 0.1));
          verify(mockOnTap).called(1);
        });
      });

      for (final AxisDirection direction in AxisDirection.values) {
        group('$direction', () {
          final Rect interactionRect = <AxisDirection, Rect>{
            AxisDirection.up:
                Offset(0, kScreenSize.height - interationZoneExtent) &
                    Size(kScreenSize.width, interationZoneExtent),
            AxisDirection.down:
                Offset.zero & Size(kScreenSize.width, interationZoneExtent),
            AxisDirection.right:
                Offset.zero & Size(interationZoneExtent, kScreenSize.height),
            AxisDirection.left:
                Offset(kScreenSize.width - interationZoneExtent, 0) &
                    Size(interationZoneExtent, kScreenSize.height),
          }[direction]!;

          group('Pointer detected inside', () {
            final Rect insideArea =
                EdgeInsets.all(0.1).deflateRect(interactionRect);
            testWidgets('at top left', (WidgetTester tester) async {
              await tester.pumpInteractionZone(
                direction: direction,
                interationZoneExtent: interationZoneExtent,
                onTap: mockOnTap,
              );
              await tester.tapAt(insideArea.topLeft);
              verify(mockOnTap).called(1);
            });

            testWidgets('at top right', (WidgetTester tester) async {
              await tester.pumpInteractionZone(
                direction: direction,
                interationZoneExtent: interationZoneExtent,
                onTap: mockOnTap,
              );
              await tester.tapAt(insideArea.topRight);
              verify(mockOnTap).called(1);
            });

            testWidgets('at center', (WidgetTester tester) async {
              await tester.pumpInteractionZone(
                direction: direction,
                interationZoneExtent: interationZoneExtent,
                onTap: mockOnTap,
              );
              await tester.tapAt(insideArea.center);
              verify(mockOnTap).called(1);
            });
            testWidgets('at bottom left', (WidgetTester tester) async {
              await tester.pumpInteractionZone(
                direction: direction,
                interationZoneExtent: interationZoneExtent,
                onTap: mockOnTap,
              );
              await tester.tapAt(insideArea.bottomLeft);
              verify(mockOnTap).called(1);
            });

            testWidgets('at bottom right', (WidgetTester tester) async {
              await tester.pumpInteractionZone(
                direction: direction,
                interationZoneExtent: interationZoneExtent,
                onTap: mockOnTap,
              );
              await tester.tapAt(insideArea.bottomRight);
              verify(mockOnTap).called(1);
            });
          });
          group('Pointer not detected outside', () {
            final Rect outsideArea =
                EdgeInsets.all(0.1).inflateRect(interactionRect);
            testWidgets('at top left', (WidgetTester tester) async {
              await tester.pumpInteractionZone(
                direction: direction,
                interationZoneExtent: interationZoneExtent,
                onTap: mockOnTap,
              );
              await tester.tapAt(outsideArea.topLeft);
              verifyNever(mockOnTap);
            });

            testWidgets('at top right', (WidgetTester tester) async {
              await tester.pumpInteractionZone(
                direction: direction,
                interationZoneExtent: interationZoneExtent,
                onTap: mockOnTap,
              );
              await tester.tapAt(outsideArea.topRight);
              verifyNever(mockOnTap);
            });

            testWidgets('at bottom left', (WidgetTester tester) async {
              await tester.pumpInteractionZone(
                direction: direction,
                interationZoneExtent: interationZoneExtent,
                onTap: mockOnTap,
              );
              await tester.tapAt(outsideArea.bottomLeft);
              verifyNever(mockOnTap);
            });

            testWidgets('at bottom right', (WidgetTester tester) async {
              await tester.pumpInteractionZone(
                direction: direction,
                interationZoneExtent: interationZoneExtent,
                onTap: mockOnTap,
              );
              await tester.tapAt(outsideArea.bottomRight);
              verifyNever(mockOnTap);
            });
          });
        });
      }
    });

    group('MinInteractionZone no detects pointer outside of zone', () {
      testWidgets('Axis done', (WidgetTester tester) async {
        await testMinInteractionZone(
          tester: tester,
          direction: AxisDirection.down,
          interationZoneExtent: 0,
          tapOffset: const Offset(20, 20),
          shouldDetectTap: false,
        );
      });
    });
  });
}

class _VoidCallback {
  void call() {}
}

class MockVoidCallback extends Mock implements _VoidCallback {}
